#ifndef HARKTOOL4_MULTIDIMALGORITHM_HPP
#define HARKTOOL4_MULTIDIMALGORITHM_HPP

#include <iterator>
#include <algorithm>
#include <numeric>
#include <utility>

namespace harktool4
{

struct MultiDimAll {} md_all;
typedef std::pair<size_t, size_t> MultiDimRange;

template <typename V>
std::pair<typename V::iterator, typename V::iterator> GetRangeIterator(V& v, const MultiDimRange& range)
{
    std::pair<typename V::iterator, typename V::iterator> result(v.begin(), v.begin());
    std::advance(result.first, range.first);
    std::advance(result.second, range.second);
    return result;
}

template <typename V>
std::pair<typename V::iterator, size_t> GetFirstIteratorAndSize(V& v, const MultiDimRange& range)
{
    std::pair<typename V::iterator, size_t> result(v.begin(), range.second - range.first);
    std::advance(result.first, range.first);
    return result;
}

/** @brief 指定されたコンテナの指定された範囲に、関数fを適用する.
 * コンテナvの範囲[first, last)の要素へのイテレータitに対し、
 * f(*it)
 * を実行する。
 * fの中で*itを変更してもよく、*itへの変更はコンテナvに反映される。
 *
 * 範囲[first, last)は任意の数だけ指定することができる。
 * 範囲が2個以上指定された場合、多次元配列に対するfor eachとみなし、
 * 指定された範囲の個数だけ再帰的に実行する。
 */
template <typename V, typename Function>
void MultiDimForEach(V& v, Function f, const MultiDimRange& range)
{
    auto range_iterator = GetRangeIterator(v, range);
    std::for_each(range_iterator.first, range_iterator.second, f);
}

template <typename V, typename Function>
void MultiDimForEach(V& v, Function f, const MultiDimAll&)
{
    std::for_each(v.begin(), v.end(), f);
}

template <typename InputIterator, typename Function, typename ... Ranges>
void MultiDimForEach(InputIterator first, Function f, size_t n, Ranges ... ranges)
{
    for (size_t i = 0; i < n; ++i, ++first)
    {
        MultiDimForEach(*first, f, ranges...);
    }
}

template <typename V, typename Function, typename ... Ranges>
void MultiDimForEach(V& v, Function f, const MultiDimRange& range, Ranges ... ranges)
{
    auto first_and_size = GetFirstIteratorAndSize(v, range);
    MultiDimForEach(first_and_size.first, f, first_and_size.second, ranges...);
}

template <typename V, typename Function, typename ... Ranges>
void MultiDimForEach(V& v, Function f, const MultiDimAll&, Ranges ... ranges)
{
    MultiDimForEach(v.begin(), f, v.size(), ranges...);
}

/** @brief 指定されたコンテナの指定された範囲に、定数valueを格納する.
 * コンテナvの範囲[first, last)の要素へのイテレータitに対し、
 * *it = value
 * を実行する。
 *
 * 範囲[first, last)は任意の数だけ指定することができる。
 * 範囲が2個以上指定された場合、多次元配列に対するfillとみなし、
 * 指定された範囲の個数だけ再帰的に実行する。
 */
template <typename V, typename T>
void MultiDimFill(V& v, const T& value, const MultiDimRange& range)
{
    auto first_and_size = GetFirstIteratorAndSize(v, range);
    std::fill_n(first_and_size.first, first_and_size.second, value);
}

template <typename V, typename T>
void MultiDimFill(V& v, const T& value, const MultiDimAll&)
{
    std::fill_n(v.begin(), v.size(), value);
}

template <typename V, typename T, typename ... Ranges>
void MultiDimFill(V& v, const T& value, const MultiDimRange& range, Ranges ... ranges);
template <typename V, typename T, typename ... Ranges>
void MultiDimFill(V& v, const T& value, const MultiDimAll&, Ranges ... ranges);

template <typename InputIterator, typename T, typename ... Ranges>
void MultiDimFill(InputIterator first, const T& value, size_t n, Ranges ... ranges)
{
    for (size_t i = 0; i < n; ++i, ++first)
    {
        MultiDimFill(*first, value, ranges...);
    }
}

template <typename V, typename T, typename ... Ranges>
void MultiDimFill(V& v, const T& value, const MultiDimRange& range, Ranges ... ranges)
{
    auto first_and_size = GetFirstIteratorAndSize(v, range);
    MultiDimFill(first_and_size.first, value, first_and_size.second, ranges...);
}

template <typename V, typename T, typename ... Ranges>
void MultiDimFill(V& v, const T& value, const MultiDimAll&, Ranges ... ranges)
{
    MultiDimFill(v.begin(), value, v.size(), ranges...);
}

/** @brief 指定されたコンテナの指定された範囲の総和を計算する.
 * コンテナvの範囲[first, last)の要素へのイテレータitに対し、
 * init_value = binary_op(init_value, *it)
 * を実行する。
 *
 * 範囲[first, last)は任意の数だけ指定することができる。
 * 範囲が2個以上指定された場合、多次元配列に対するaccumulateとみなし、
 * 指定された範囲の個数だけ再帰的に実行する。
 */
template <typename V, typename T, typename BinaryOperator>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimRange& range)
{
    auto range_iterator = GetRangeIterator(v, range);
    return std::accumulate(range_iterator.first, range_iterator.second, init, binary_op);
}

template <typename V, typename T, typename BinaryOperator>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimAll&)
{
    return std::accumulate(v.begin(), v.end(), init, binary_op);
}

template <typename V, typename T, typename BinaryOperator, typename ... Ranges>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimRange& range, Ranges ... ranges);
template <typename V, typename T, typename BinaryOperator, typename ... Ranges>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimAll&, Ranges ... ranges);

template <typename InputIterator, typename T, typename BinaryOperator, typename ... Ranges>
T MultiDimAccumulate(InputIterator first, const T& init, BinaryOperator binary_op, size_t n, Ranges ... ranges)
{
    T temp = init;
    for (size_t i = 0; i < n; ++i, ++first)
    {
        temp = MultiDimAccumulate(*first, temp, binary_op, ranges...);
    }
    return temp;
}

template <typename V, typename T, typename BinaryOperator, typename ... Ranges>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimRange& range, Ranges ... ranges)
{
    auto first_and_size = GetFirstIteratorAndSize(v, range);
    return MultiDimAccumulate(first_and_size.first, init, binary_op, first_and_size.second, ranges...);
}

template <typename V, typename T, typename BinaryOperator, typename ... Ranges>
T MultiDimAccumulate(V& v, const T& init, BinaryOperator binary_op, const MultiDimAll&, Ranges ... ranges)
{
    return MultiDimAccumulate(v.begin(), init, binary_op, v.size(), ranges...);
}

} // harktool4

#endif // HARKTOOL4_MULTIDIMALGORITHM_HPP
